/***************************************************************************

	cmdlist.c

	MAME Plus! format command list viewer.

***************************************************************************/

#include "ncdz.h"

#define issjis1(c)	(((c) >= 0x81 && (c) <= 0x9f) | ((c) >= 0xe0 && (c) <= 0xfc))
#define issjis2(c)	((c) >= 0x40 && (c) <= 0xfc && (c) != 0x7f)

#define CR		0x0d
#define LF		0x0a
#define CRLF	"\x0d\x0a"

enum
{
	LF_WIN = 0,
	LF_MAC,
	LF_UNIX,
	LF_MAX
};

enum
{
	TAG_INFO = 0,
	TAG_CMD,
	TAG_END,
	TAG_CHARSET,
	TAG_MAX
};


/***************************************************************************
	Private structures
 ***************************************************************************/

typedef struct cmdlist_t
{
	int lines;
	char **line;
} CMDLIST;

static CMDLIST **cmd;


/***************************************************************************
	Private variables
 ***************************************************************************/

static const char *tag_str[TAG_MAX] =
{
	"info",
	"cmd",
	"end",
	"charset"
};

static int lf_code;
static int charset;

static char *cmdbuf;
static char **cmdline;
static int num_items, show_items, rows_item, sel_item, prev_item, top_item;
static int num_lines, show_lines, rows_line, sel_line, prev_line;
static int item_min_sx, item_max_sx;


/*--------------------------------------------------------
	Cobt@^O擾
 -------------------------------------------------------*/

static int cmdlist_get_tag(char *buf)
{
	int i;

	for (i = 0; i < TAG_MAX; i++)
		if (strnicmp(&buf[1], tag_str[i], strlen(tag_str[i])) == 0)
			return i;

	return -1;
}


/*--------------------------------------------------------
	Cobt@l擾
 -------------------------------------------------------*/

static char *cmdlist_get_value(char *buf)
{
	if (strtok(buf, " =\r\n") != NULL)
		return strtok(NULL, " =\r\t");

	return NULL;
}


/*--------------------------------------------------------
	eLXg̉sR[h𔻕
 -------------------------------------------------------*/

static int check_linefeed_code(const char *buf)
{
	if (strstr(buf, CRLF) != NULL)
	{
		// Windows / MS-DOS
		return LF_WIN;
	}
	else if (strchr(buf, CR) != NULL)
	{
		// Mac
		return LF_MAC;
	}

	// Unix
	return LF_UNIX;
}


/*--------------------------------------------------------
	eLXg̕R[h𔻕
 -------------------------------------------------------*/

#if JAPANESE_UI
static int check_text_encode(char *buf, int size)
{
	int i;
	int count1, count2, per;
	u8 *p = (u8 *)buf;

	count1 = 0;
	count2 = 0;

	for (i= 0; i < size;)
	{
		if (p[i] <= 0x20)
		{
			count2++;
			i++;
		}
		else if (p[i] == 0x97 && p[i + 1] == 0x97)
		{
			count2 += 2;
			i += 2;
		}
		else if (issjis1(p[i]) && issjis2(p[i + 1]))
		{
			count1++;
			i += 2;
		}
		else i++;
	}

	per = (count1 * 100) / (size - count2);

	if (per > 20)
		return CHARSET_SHIFTJIS;
	else
		return CHARSET_LATIN1;
}
#endif


/*--------------------------------------------------------
	R}hXg̃[h
 -------------------------------------------------------*/

void load_commandlist(const char *game_name)
{
	FILE *fp;
	char path[MAX_PATH];
	char *p, linebuf[256];
	int i, found, line, item;
	int size, start, end, pos;

	cmdbuf    = NULL;
	cmdline   = NULL;
	cmd       = NULL;
	charset   = CHARSET_DEFAULT;
	num_lines = 0;
	num_items = 0;

	sprintf(path, "%sdata/command.dat", launchDir);

	if ((fp = fopen(path, "rb")) == NULL)
		return;

	if (fgets(linebuf, 255, fp) == NULL)
	{
		// ̃t@C
		fclose(fp);
		return;
	}

	// sR[h`FbN
	lf_code = check_linefeed_code(linebuf);

	// R}h͊Jn
	fseek(fp, 0, SEEK_SET);
	found = 0;
	line  = 0;
	start = 0;
	end   = 0;
	pos   = 0;
	while (fgets(linebuf, 255, fp) != NULL)
	{
		if (linebuf[0] == '$')
		{
			switch (cmdlist_get_tag(linebuf))
			{
			case TAG_INFO:
				if (found && start && end)
				{
					// I
					found = 2;
				}
				else if (found)
				{
					// R}hAُ
					found = 0;
				}
				else if ((p = cmdlist_get_value(linebuf)) != NULL)
				{
					if (stricmp(p, game_name) == 0)
					{
						// R}h
						found = 1;
					}
				}
				break;

			case TAG_CMD:
				if (found)
				{
					if (!start) start = pos;
					num_items++;
				}
				break;

			case TAG_END:
				if (start)
				{
					end = ftell(fp);
					num_lines = line + 1;
				}
				break;

#if JAPANESE_UI
			case TAG_CHARSET:
				if ((p = cmdlist_get_value(linebuf)) != NULL)
				{
					if (stricmp(p, "Shift_JIS") == 0)
					{
						charset = CHARSET_SHIFTJIS;
					}
					else if (stricmp(p, "ISO-8859-1") == 0)
					{
						charset = CHARSET_ISO8859_1;
					}
					else if (stricmp(p, "Latin1") == 0)
					{
						charset = CHARSET_ISO8859_1;
					}
				}
				break;
#endif
			}

			if (found == 2) break;
		}

		if (start) line++;
		else pos = ftell(fp);
	}

	if (!found)
	{
		// Ȃꍇ͏I
		fclose(fp);
		return;
	}

	// obt@mۂAǂݍ
	size = end - start;

	if ((cmdbuf = calloc(1, size)) == NULL)
	{
		fclose(fp);
		return;
	}

	fseek(fp, start, SEEK_SET);
	fread(cmdbuf, 1, size, fp);
	fclose(fp);

	// R[h`FbN
#if JAPANESE_UI
	if (charset == CHARSET_DEFAULT)
		charset = check_text_encode(cmdbuf, size);
#endif

	// spm
	if ((cmdline = calloc(num_lines, sizeof(char *))) == NULL)
		goto error;

	// obt@sɕ
	p = cmdbuf;
	for (line = 0; line < num_lines; line++)
	{
		cmdline[line] = p;

		switch (lf_code)
		{
		case LF_WIN:
			p = strstr(cmdline[line], CRLF);
			*p++ = '\0';
			*p++ = '\0';
			break;

		case LF_MAC:
			p = strchr(cmdline[line], CR);
			*p++ = '\0';
			break;

		case LF_UNIX:
			p = strchr(cmdline[line], LF);
			*p++ = '\0';
			break;
		}
	}

	// R}hXg\̂̊m
	if ((cmd = (CMDLIST **)calloc(num_items, sizeof(CMDLIST *))) == NULL)
		goto error;

	for (i = 0; i < num_items; i++)
	{
		if ((cmd[i] = (CMDLIST *)calloc(1, sizeof(CMDLIST))) == NULL)
			goto error;
	}

	// eڂ̍s`FbNAo^
	for (i = 0; i < 2; i++)
	{
		int item_line = 0;

		line  = 0;
		item  = 0;
		found = 0;

		while (line < num_lines && item < num_items)
		{
			strcpy(linebuf, cmdline[line]);

			if (linebuf[0] == '$')
			{
				switch (cmdlist_get_tag(linebuf))
				{
				case TAG_CMD:
					found = 1;
					item_line = 0;
					break;

				case TAG_END:
					found = 0;
					item++;
					break;
				}
			}
			else if (found)
			{
				if (i == 0)
					cmd[item]->lines++;
				else
					cmd[item]->line[item_line++] = cmdline[line];
			}
			line++;
		}

		if (i == 0)
		{
			for (item = 0; item < num_items; item++)
			{
				if ((cmd[item]->line = calloc(1, sizeof(char *) * cmd[item]->lines)) == NULL)
					goto error;
			}
		}
	}

	free(cmdline);
	cmdline = NULL;

	// \̏
	sel_line   = 0;
	prev_line  = 0;
#if JAPANESE_UI
	rows_line  = (charset & CHARSET_SHIFTJIS) ? 16 : 14;
#else
	rows_line  = 14;
#endif
	show_lines = rows_line;
	num_lines  = cmd[0]->lines;
	if (num_lines < show_lines) show_lines = num_lines;

	top_item   = 0;
	sel_item   = 0;
	prev_item  = 0;
	rows_item  = 13;
	show_items = rows_item;
	if (num_items < show_items) show_items = num_items;

	// ڃj[̕vZ
	item_min_sx = 12 + 46 * 7;
	item_max_sx = 480;
	for (item = 0; item < num_items; item++)
	{
		int x;

		x = 480 - (strlen(cmd[item]->line[0]) * 7 + 16);
		if (item_max_sx > x) item_max_sx = x;

		for (line = 0; line < cmd[item]->lines; line++)
		{
			x = 12 + strlen(cmd[item]->line[line]) * 7;
			if (item_min_sx < x) item_min_sx = x;
		}
	}

	return;

error:
	free_commandlist();
}


/*--------------------------------------------------------
	R}hXg̉
 -------------------------------------------------------*/

void free_commandlist(void)
{
	int i;

	if (cmd)
	{
		for (i = 0; i < num_items; i++)
		{
			if (cmd[i])
			{
				if (cmd[i]->line) free(cmd[i]->line);
				free(cmd[i]);
			}
		}
		free(cmd);
		cmd = NULL;
	}

	if (cmdline)
	{
		free(cmdline);
		cmdline = NULL;
	}

	if (cmdbuf)
	{
		free(cmdbuf);
		cmdbuf = NULL;
	}
}


/*--------------------------------------------------------
	R}hXg\
 -------------------------------------------------------*/

void commandlist(int flag)
{
	int x, y, alpha;
	int update = 1, menu_open = 0, menu_counter = 0;
	int mp3_paused = 0;
	char title[80], temp[32];
	extern const char *game_name;

	if (cmdbuf == NULL)
	{
#if JAPANESE_UI
		ui_popup("̃Q[p̃R}hXg܂B");
#else
		ui_popup("Command list for this game not found.");
#endif
		return;
	}

	if (flag)
	{
		if (mp3_get_status() == MP3_PLAY)
		{
			mp3_pause(1);
			mp3_paused = 1;
		}
		sound_enable(0);
		video_set_mode(32);
		set_cpu_clock(PSPCLOCK_222);
	}

	pad_wait_clear();
	load_background(WP_CMDLIST);
	ui_popup_reset();

#if JAPANESE_UI
	sprintf(title, "R}hXg - %s", game_name);
#else
	sprintf(title, "Command list - %s", game_name);
#endif

	do
	{
		if (update)
		{
			update = 0;

			show_background();

			small_icon_shadow(8, 3, UI_COLOR(UI_PAL_TITLE), ICON_CMDLIST);
			uifont_print_shadow(36, 5, UI_COLOR(UI_PAL_TITLE), title);
			draw_battery_status(1);

			if (sel_line != 0)
				uifont_print_shadow(161, 24, UI_COLOR(UI_PAL_SELECT), FONT_UPTRIANGLE);

#if JAPANESE_UI
			if (charset & CHARSET_SHIFTJIS)
			{
				for (y = 0; y < show_lines; y++)
					textfont_print(6, 37 + 14 * y, UI_COLOR(UI_PAL_SELECT), cmd[sel_item]->line[y + sel_line], charset);
			}
			else
			{
				for (y = 0; y < show_lines; y++)
					textfont_print(6, 37 + 16 * y, UI_COLOR(UI_PAL_SELECT), cmd[sel_item]->line[y + sel_line], charset);
			}
#else
			for (y = 0; y < show_lines; y++)
				textfont_print(6, 37 + 16 * y, UI_COLOR(UI_PAL_SELECT), cmd[sel_item]->line[y + sel_line], CHARSET_ISO8859_1);
#endif

			if (sel_line + show_lines < num_lines)
				uifont_print_shadow(161, 260, UI_COLOR(UI_PAL_SELECT), FONT_DOWNTRIANGLE);

			if (menu_open)
			{
				alpha = 14;
				x = item_max_sx;
			}
			else if (menu_counter > 0)
			{
				alpha = 14 - ((4 - menu_counter) << 1);
				x = item_max_sx + ((item_min_sx - item_max_sx) >> 2) * (4 - menu_counter);
			}
			else
			{
				alpha = 6;
				x = item_min_sx;
			}

			boxfill_alpha(x, 26, 479, 271, UI_COLOR(UI_PAL_BG1), alpha);

			for (y = 0; y < rows_item; y++)
			{
				if (top_item + y >= num_items) break;

#if JAPANESE_UI
				if (top_item + y == sel_item)
				{
					uifont_print(x, 37 + 16 * y, UI_COLOR(UI_PAL_SELECT), FONT_RIGHTTRIANGLE);
					textfont_print(x + 14, 37 + 16 * y, UI_COLOR(UI_PAL_SELECT), cmd[top_item + y]->line[0], charset);
				}
				else
					textfont_print(x + 14, 37 + 16 * y, UI_COLOR(UI_PAL_NORMAL), cmd[top_item + y]->line[0], charset);
#else
				textfont_print(x + 14, 37 + 16 * y, UI_COLOR(UI_PAL_NORMAL), cmd[top_item + y]->line[0], CHARSET_ISO8859_1);
#endif
			}

#if JAPANESE_UI
			sprintf(temp, "%d/%d ", sel_item + 1, num_items);
#else
			sprintf(temp, "%d/%d items", sel_item + 1, num_items);
#endif
			x = uifont_get_string_width(temp);
			uifont_print(475 - x, 250, UI_COLOR(UI_PAL_SELECT), temp);

			update |= ui_show_popup(1);
			video_flip_screen(1);
		}
		else
		{
			update = ui_show_popup(0);
			video_wait_vsync();
		}

		if (menu_counter)
		{
			update = 1;
			menu_counter--;
		}

		prev_item = sel_item;
		prev_line = sel_line;
		pad_update();

		if (pad_pressed(PSP_CTRL_UP))
		{
			if (menu_open)
			{
				if (sel_item > 0)
				{
					sel_item--;
					show_lines = rows_line;
					num_lines = cmd[sel_item]->lines;
					if (num_lines < show_lines)
						show_lines = num_lines;
					sel_line = 0;
				}
			}
			else
			{
				if (sel_line > 0) sel_line--;
			}
		}
		else if (pad_pressed(PSP_CTRL_DOWN))
		{
			if (menu_open)
			{
				if (sel_item < num_items - 1)
				{
					sel_item++;
					show_lines = rows_line;
					num_lines = cmd[sel_item]->lines;
					if (num_lines < show_lines)
						show_lines = num_lines;
					sel_line = 0;
				}
			}
			else
			{
				if (sel_line + show_lines < num_lines) sel_line++;
			}
		}
		else if (pad_pressed(PSP_CTRL_LEFT))
		{
			if (sel_line > 0)
			{
				sel_line -= rows_line;
				if (sel_line < 0) sel_line = 0;
			}
		}
		else if (pad_pressed(PSP_CTRL_RIGHT))
		{
			if (sel_line + show_lines < num_lines)
			{
				sel_line += rows_line;
				if (sel_line + show_lines > num_lines)
					sel_line = num_lines - show_lines;
			}
		}
		else if (pad_pressed(PSP_CTRL_CIRCLE))
		{
			menu_open ^= 1;
			update = 1;
			if (menu_open == 0) menu_counter = 4;
			pad_wait_clear();
		}
		else if (pad_pressed(PSP_CTRL_START))
		{
			ui_save_screenshot("menu6");
			pad_wait_clear();
		}
		else if (pad_pressed(PSP_CTRL_RTRIGGER))
		{
			help(HELP_COMMANDLIST);
			update = 1;
		}

		if (menu_open)
		{
			if (top_item > num_items - rows_item) top_item = num_items - rows_item;
			if (top_item < 0) top_item = 0;
			if (sel_item >= top_item + rows_item) top_item = sel_item - rows_item + 1;
			if (sel_item < top_item) top_item = sel_item;
		}

		if ((sel_line != prev_line) || (sel_item != prev_item))
			update = 1;

	} while (!pad_pressed(PSP_CTRL_CROSS));

	pad_wait_clear();
	ui_popup_reset();

	if (flag)
	{
		set_cpu_clock(psp_cpuclock);

		video_set_mode(16);
		init_autoframeskip();
		blit_clear_all_cache();

		sound_set_volume();
		sound_enable(1);

		mp3_file_reopen();
		mp3_set_volume();
		if (mp3_paused) mp3_pause(0);
	}
	else
	{
		load_background(WP_LOGO);
	}
}


#if JAPANESE_UI

#define issjis1(c)	(((c) >= 0x81 && (c) <= 0x9f) | ((c) >= 0xe0 && (c) <= 0xfc))
#define issjis2(c)	((c) >= 0x40 && (c) <= 0xfc && (c) != 0x7f)

#define MAX_LINES	13
#define MIN_X		24
#define MIN_Y		47
#define INC_Y		16

#define INFO_SEEK	0
#define CMD_SEEK	1
#define END_SEEK	2

static int cy;
static char msg_lines[MAX_LINES][256];


static void msg_printf(const char *text, ...)
{
	int y;
	char buf[256];
	va_list arg;

	va_start(arg, text);
	vsprintf(buf, text, arg);
	va_end(arg);

	if (cy == MAX_LINES - 1)
	{
		for (y = 1; y < MAX_LINES; y++)
			strcpy(msg_lines[y - 1], msg_lines[y]);
	}

	strcpy(msg_lines[cy], buf);

	show_background();
	draw_battery_status(1);

	for (y = 0; y <= cy; y++)
		uifont_print(MIN_X, MIN_Y + y * 16, COLOR_WHITE, msg_lines[y]);

	if (cy < MAX_LINES - 1) cy++;

	video_flip_screen(1);
}


int reduce_commandlist(void)
{
	FILE *fp;
	char path[MAX_PATH], path2[MAX_PATH];
	char *p, linebuf[256], temp[256];
	int i, j, l, found;
	int num_games, charset;
	int header_end, body_start, body_end;
	int progress, charset2;
	int line = 0, line2 = 0, num_cmd;
	int org_size, new_size, comment;
	char *textbuf = NULL, **line_ptr = NULL;

	sprintf(path, "%sdata/command.dat", launchDir);
	if ((fp = fopen(path, "rb")) == NULL)
		return 0;

	pad_wait_clear();
	load_background(WP_CMDLIST);
	ui_popup_reset();

	show_background();
	small_icon_shadow(6, 3, UI_COLOR(UI_PAL_TITLE), ICON_COMMANDDAT);
	uifont_print_shadow(32, 5, UI_COLOR(UI_PAL_TITLE), "COMMAND.DAT TCYk");
	draw_dialog(14, 37, 465, 259);
	video_copy_rect32(draw_frame, work_frame, &full_rect, &full_rect);

	cy = 0;

	num_games  = 0;
	header_end = -1;
	body_start = -1;
	body_end   = -1;
	charset    = CHARSET_DEFAULT;

	msg_printf("{^ƏJn܂B");
	pad_wait_press(-1);
	pad_wait_clear();

	fseek(fp, 0, SEEK_END);
	org_size = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	msg_printf("\n");
	msg_printf("COMMAND.DAT ...");

	// o^ĂQ[`FbN
	while (fgets(linebuf, 255, fp) != NULL)
	{
		if (strrchr(linebuf, '\r') == NULL)
		{
			if (strrchr(linebuf, '\n') == NULL)
			{
				msg_printf("(ERROR) line %d: 1s̕܂B(ő255)", line + 1);
				fclose(fp);
				goto error;
			}
		}

		if (linebuf[0] == '$')
		{
			p = strtok(linebuf, "\r\n");

			if (p && linebuf[0] == '$')
			{
				if (!strnicmp(linebuf, "$charset", 8) && strchr(linebuf, '=') != NULL)
				{
					char *tag  = strtok(linebuf, " =");
					char *type = strtok(NULL, " =");

					if (type)
					{
						if (!stricmp(type, "Shift_JIS"))
						{
							charset = CHARSET_SHIFTJIS;
							msg_printf("line %d: Zbg Shift_JIS w肳܂B", line + 1);
						}
						else if (!stricmp(type, "ISO-8859-1") || !stricmp(type, "Latin1"))
						{
							charset = CHARSET_LATIN1;
							msg_printf("line %d: Zbg ISO-8859-1 w肳܂B", line + 1);
						}
					}
				}
				else if (!strncmp(linebuf, "$info", 5) && strchr(linebuf, '=') != NULL)
				{
					char *tag  = strtok(linebuf, " =");
					char *name = strtok(NULL, " =");

					if (name)
					{
						if ((p = strchr(name, ',')) != NULL)
						{
							*p = '\0';
							strcat(name, "...");
						}

						if (body_start == -1)
						{
							// ŏ $info L^
							body_start = line;
						}
						num_games++;
						found = 1;
					}
				}
				else if (!strncmp(linebuf, "$end", 4))
				{
					// Ō $end L^
					body_end = line;
				}
			}
		}
		else if (body_start == -1)
		{
			if (!strcmp(linebuf, "\r\n")) header_end = line;
		}

		line++;
	}

	msg_printf("t@ČI܂B");

	if (!found)
	{
		msg_printf("(ERROR) Ή\"COMMAND.DAT\"t@Č`قȂ܂B");
		fclose(fp);
		goto error;
	}
	if (line == 0)
	{
		msg_printf("(ERROR) ̃t@CłB");
		fclose(fp);
		goto error;
	}

	msg_printf("\n");
	msg_printf("s %d", line);
	msg_printf("o^Q[ %d", num_games);
	msg_printf("\n");
	msg_printf("t@CTCYkJn܂B");

	if ((line_ptr = (char **)malloc(sizeof(char *) * line)) == NULL)
	{
		msg_printf("(ERROR) mۂł܂B");
		goto error;
	}
	if ((textbuf = (char *)malloc(org_size + 1)) == NULL)
	{
		msg_printf("(ERROR) mۂł܂B");
		goto error;
	}
	memset(textbuf, 0, org_size + 1);

	msg_printf("t@Cǂݍݒ...");
	fseek(fp, 0, SEEK_SET);
	fread(textbuf, 1, org_size, fp);
	fclose(fp);

	if (charset == CHARSET_DEFAULT)
	{
		charset2 = check_text_encode(textbuf, org_size);

		if (charset2 == CHARSET_SHIFTJIS)
			msg_printf("{ŃR}hXgƔʂ܂Ał?");
		else
			msg_printf("pŃR}hXgƔʂ܂Ał?");
		msg_printf("͂A~ ĂB");

		do
		{
			video_wait_vsync();
			pad_update();

			if (pad_pressed(PSP_CTRL_CIRCLE))
			{
				charset = charset2;
				pad_wait_clear();
				break;
			}

		} while (!pad_pressed(PSP_CTRL_CROSS));
	}

	// s̈ʒuۑ
	line_ptr[0] = strtok(textbuf, "\n");

	i = 1;
	while (1)
	{
		if ((p = strtok(NULL, "\n")) == NULL) break;
		line_ptr[i++] = p;
	}

	for (i = 0; i < line; i++)
	{
		if ((p = strrchr(line_ptr[i], '\r')) != NULL)
			*p = '\0';
	}

	// ޔt@C쐬
	sprintf(path2, "%sdata/command.org", launchDir);

	msg_printf("\n");
	msg_printf("̃t@Cl[܂B");
	sceIoRemove(path2);

	// l[đޔ (ޔt@C: command.org)
	if (sceIoRename(path, path2) < 0)
	{
		msg_printf("(ERROR) t@Cl[ł܂łB");
		goto error;
	}

	//------------------------------------------------------------------
	// kJn
	//------------------------------------------------------------------
	if ((fp = fopen(path, "w")) == NULL)
	{
		msg_printf("(ERROR) o̓t@CI[vł܂łB");
		goto error;
	}

	l = 0;
	line2 = 0;
	num_cmd = 0;
	progress = INFO_SEEK;

	msg_printf("\n");

	if (charset != CHARSET_DEFAULT)
	{
		msg_printf("Zbgw^Oo͂܂B");

		if (charset == CHARSET_SHIFTJIS)
			fprintf(fp, "$charset=shift_jis\r\n");
		else
			fprintf(fp, "$charset=latin1\r\n");

		line2++;
	}

	// wb_Rs[
	if (header_end != -1)
	{
		msg_printf("wb_Rs[...");

		for (; l <= header_end; l++)
		{
			strcpy(linebuf, line_ptr[l]);

			if (strnicmp(linebuf, "$charset", 8) != 0)
			{
				fprintf(fp, "%s\r\n", linebuf);
				line2++;
			}
		}

		msg_printf("wb_̃Rs[");
	}

	msg_printf("R}hXgf[^̃Rs[Jn");

	for (; l <= body_end; l++)
	{
		strcpy(linebuf, line_ptr[l]);

		switch (progress)
		{
		case INFO_SEEK:
			if (linebuf[0] == '$')
			{
				// R}hXgJn
				if (!strnicmp(linebuf, "$info", 5) && strchr(linebuf, '=') != NULL)
				{
					char *tag  = strtok(linebuf, " =");
					char *name = strtok(NULL, " =");

					if (name)
					{
						found = 0;

						if (!stricmp(name, "trally")) strcpy(name, "rallych");

						for (i = 0; i < 97; i++)
						{
							if (!stricmp(name, games[i].name))
							{
								found = 1;
								break;
							}
						}

						if (found)
						{
							if (l != 0)
							{
								j = l;
								while (j > body_start)
								{
									if (line_ptr[j - 1][0] == '#') j--;
									else break;
								}

								while (j < l)
								{
									fprintf(fp, "%s\r\n", line_ptr[j]);
									j++;
								}
							}

							msg_printf("%s ̃R}hRs[...", games[i].name);
							fprintf(fp, "$info=%s\r\n", games[i].name);

							progress = CMD_SEEK;
							num_cmd++;
							line2++;
						}
					}
				}
			}
			break;

		case CMD_SEEK:
			if (!strnicmp(linebuf, "$cmd", 4))
			{
				// R}hJn
				progress = END_SEEK;
				fprintf(fp, "$cmd\r\n");
				line2++;
			}
			else if (!strnicmp(linebuf, "$info", 5))
			{
				// ̃R}h - 1s߂
				fprintf(fp, "\r\n");
				progress = INFO_SEEK;
				l--;
			}
			break;

		case END_SEEK:
			if (!strnicmp(linebuf, "$end", 4))
			{
				// R}hI
				progress = CMD_SEEK;
				fprintf(fp, "$end\r\n");
				line2++;
			}
			else
			{
				// R}hXg̒g - ̂܂܏o
				fprintf(fp, "%s\r\n", linebuf);
				line2++;
			}
			break;
		}
	}

	msg_printf("R}hXgf[^̃Rs[");

	if (body_end < line - 1)
	{
		msg_printf("tb^Rs[...");

		if (strlen(line_ptr[l]) == 0)
			l++;

		for (; l < line; l++)
		{
			strcpy(linebuf, line_ptr[l]);
			fprintf(fp, "%s\r\n", linebuf);
			line2++;
		}

		msg_printf("tb^̃Rs[");
	}

	fclose(fp);

	fp = fopen(path, "rb");
	fseek(fp, 0, SEEK_END);
	new_size = ftell(fp);
	fclose(fp);

	msg_printf("\n");
	msg_printf("ϊ s %d", line2);
	msg_printf("ϊ o^Q[ %d", num_cmd);
	msg_printf("TCY %dbytes  %dbytes (%.1f%%)", org_size, new_size, 100.0 - ((float)new_size / (float)org_size) * 100.0);

	msg_printf("\n");
	msg_printf("܂B");

error:
	if (textbuf) free(textbuf);
	if (line_ptr) free(line_ptr);

	msg_printf("\n");
	msg_printf("{^ĂB");

	pad_wait_press(-1);
	pad_wait_clear();
	ui_popup_reset();

	load_background(WP_FILER);
	return 1;
}

#endif
